-- Attack (0-100%) - Decay (100-70%) -- Sustain (70%) -- Release (70-0%) (ADSR) module
-- made up of 
-- a finite-state machine to control the ADSR sequence
-- a time counter to measure time
-- a gain that varies linearily and will be multiplied by the tone

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.MATH_REAL.ALL;

entity ADSR_module is
    generic (
        N: natural := 8; -- number of bits in the output signal
        N_t: natural := 16; --number of bits to count the time
        A_t: real := 20.0; -- length of the attack in ms
        D_t: real := 5.0; -- length of the decay in ms
        R_t: real := 700.0; -- length of the release in ms
        fs: real := 48000.0 -- sampling frequency
    );
    port (
        clk: in std_logic;
        rst: in std_logic;
        key: in std_logic ; -- key input
        gain: out std_logic_vector(N-1 downto 0) -- unsigned value
    );
end ADSR_module;

architecture archi of ADSR_module is

        -- constants for the time calculations
        constant A_cycles: unsigned(N_t-1 downto 0) := to_unsigned(integer(A_t * fs / 1000.0), N_t);
        constant D_cycles: unsigned(N_t-1 downto 0) := to_unsigned(integer(D_t * fs / 1000.0), N_t);
        constant R_cycles: unsigned(N_t-1 downto 0) := to_unsigned(integer(R_t * fs / 1000.0), N_t);

        constant A_coeff : unsigned(N_t-1 downto 0) := to_unsigned(integer(real(2**N-1)/(A_t * fs / 1000.0)*real(2**N_t-1)), N_t);
        constant D_coeff : unsigned(N_t-1 downto 0) := to_unsigned(integer(0.3*real(2**N-1)/(D_t * fs / 1000.0)*real(2**N_t-1)), N_t);
        constant R_coeff : unsigned(N_t-1 downto 0) := to_unsigned(integer(real(2**N-1)/(R_t * fs / 1000.0)*real(2**N_t-1)), N_t);

-- states of the FSM 
-- key_off, attack, decay, sustain, release
type state_type is (key_off, attack, decay, sustain, release_st );
signal state: state_type := key_off;
signal counter: unsigned(N_t-1 downto 0) := (others =>'0'); -- time counter
signal gain_u : signed(2*N_t downto 0) := (others =>'0');

begin

    -- When the key is pressed, the cycles begins
    process(clk, rst)
        -- counter for the time
    begin
        if rst = '1' then
            state <= key_off;
            gain_u <= (others => '0');
            counter <= (others =>'0');
        elsif rising_edge(clk) then
            case state is
                when key_off =>
                    if key = '1' then
                        state <= attack;
                        counter <= (others =>'0');
                    end if;
                when attack =>
                    if counter < A_cycles then
                        gain_u <= signed('0'&(counter*A_coeff));
                        counter <= counter + 1;
                    else
                        state <= decay;
                        counter <= (others =>'0');
                    end if;
                when decay =>
                    if counter < D_cycles then
                        gain_U <= signed('0'&(A_cycles*A_coeff)) - signed('0'&(counter * D_coeff));
                        counter <= counter + 1;
                    else
                        state <= sustain;
                    end if;
                when sustain =>
                    if key = '0' then
                        state <= release_st;
                        counter <= (others =>'0');
                    end if;
                when release_st =>
                    if counter < R_cycles then
                        gain_u <= signed('0'&(A_cycles*A_coeff)) - signed('0'&(D_cycles * D_coeff)) - signed('0'&(R_cycles*counter));
                        counter <= counter + 1;
                    else
                        state <= key_off;
                        gain_u <= (others => '0');
                    end if;
            end case;
        end if;
    end process;

    gain <= std_logic_vector(gain_u(N_t+N-1 downto N_t)) when gain_u>0 else (others => '0');
end archi ; 
