library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity synthesizer is
    port (clk : in std_logic;
          rst : in std_logic;

          wave_sel  : in  std_logic_vector(1 downto 0);
          config    : in  std_logic_vector(3 downto 0);  -- configuration of the tone generator
          keycode   : in  std_logic_vector (15 downto 0);
          audio_out : out std_logic_vector (23 downto 0));
end synthesizer;

architecture Behavioral of synthesizer is

    constant N : natural := 11;

    type tone_array_t is array (0 to N-1) of std_logic_vector (23 downto 0);
    type keycode_array_t is array (0 to N-1) of std_logic_vector (7 downto 0);
    type freq_array_t is array (0 to N-1) of real;

    signal tone_out   : tone_array_t;
    signal sum        : tone_array_t;
    signal key        : std_logic_vector (0 to N-1);
    constant keycodes : keycode_array_t := (X"15", X"1D", X"24", X"2D", X"2C", X"35", X"3C", X"43", X"44", X"4D", X"54");  -- 35 is for Y
    constant freqs    : freq_array_t    := (392.0, 415.0, 440.0, 466.0, 494.0, 523.0, 554.0, 587.0, 622.0, 659.0, 698.0);
                                                                                                                           --Sol   Sol#   La     Si-b    Si     Do     Do#    Re     Re#    Mi      Fa

begin

    -- key(0) <= '1' when (keycode(7 downto 0)=X"15" and keycode(15 downto 8)/=X"F0") else '0';
    -- key(1) <= '1' when (keycode(7 downto 0)=X"1D" and keycode(15 downto 8)/=X"F0") else '0';
    -- key(2) <= '1' when (keycode(7 downto 0)=X"24" and keycode(15 downto 8)/=X"F0") else '0';
    -- key(3) <= '1' when (keycode(7 downto 0)=X"2D" and keycode(15 downto 8)/=X"F0") else '0';
    -- key(4) <= '1' when (keycode(7 downto 0)=X"2C" and keycode(15 downto 8)/=X"F0") else '0';

    -- I want to be able to play two tones at the same time, and with different durations
    -- so a tone can be held while I play many other tones

    process(clk)
    begin
        if rising_edge(clk) then
            if rst = '1' then
                key <= (others => '0');
            else
                testkeys : for i in 0 to N-1 loop
                    if (keycode(7 downto 0) = keycodes(i) and keycode(15 downto 8) /= X"F0") then
                        key(i) <= '1';
                    else
                        if (keycode(7 downto 0) = keycodes(i) and keycode(15 downto 8) = X"F0") then
                            key(i) <= '0';
                        end if;
                    end if;
                end loop;
            end if;
        end if;
    end process;

    tone_gen : for i in 0 to N-1 generate
        tone_inst : entity work.tone_generator
            generic map(
                N         => 24,
                f0        => freqs(i),
                harmonics => 4,
                fs        => 48000.0)
            port map (clk => clk, rst => rst, wave_sel => wave_sel, config => config, key => key(i), tone_out => tone_out(i));

    end generate;

    -- Mix by adding the tones
    sum(0) <= std_logic_vector(signed(tone_out(0)));

    sum_gen : for i in 1 to N-1 generate
        sum(i) <= std_logic_vector(signed(sum(i-1))+signed(tone_out(i)));
    end generate;

    audio_out <= sum(N-1);

end Behavioral;
