-- VHDL module for a tone generator
-- based on waves

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;
use IEEE.MATH_REAL.all;

entity tone_generator is
    generic (
        N         : natural := 24;                              -- number of bits in the output signal
        f0        : real    := 1000.0;                          -- frequency of the tone
        harmonics : natural := 4;                               -- number of harmonics
        fs        : real    := 48000.0                          -- sampling frequency
        );
    port (
        clk      : in  std_logic;
        rst      : in  std_logic;
        wave_sel : in  std_logic_vector(1 downto 0);            -- type of waveform (sine, square, triangle, saw-tooth)
        config   : in  std_logic_vector(harmonics-1 downto 0);  -- configuration of the tone generator
        key      : in  std_logic;                               -- key input
        tone_out : out std_logic_vector(N-1 downto 0)
        );
end tone_generator;

architecture arch of tone_generator is
    -- when adding sine_waves, we have to deal with the fact that the magnitude of the sum may overflow
    -- we divide the magnitude of each harmonic by 2^i, where i is the harmonic number
    -- this way, the sum of the harmonics will not overflow

    type wawe_array_t is array (0 to harmonics-1) of std_logic_vector(N-1-natural(log2(real(harmonics))) downto 0);
    type wawe_array_sum_t is array (0 to harmonics-1) of signed(N-1 downto 0);

    signal wave_array      : wawe_array_t;
    signal wave_array_sum  : wawe_array_sum_t;
    signal gain            : std_logic_vector(7 downto 0);  -- triggered by the key
    signal tone_out_signed : signed(N-1 downto 0);
    signal tone_out_mult   : signed(N+8 downto 0);

begin

    adsr_inst : entity work.ADSR_module
        generic map (
            N   => 8,
            N_t => 16,                  --number of bits to count the time
            A_t => 10.0,                -- length of the attack in ms
            D_t => 2.0,                 -- length of the decay in ms
            R_t => 50.0,                -- length of the release in ms
            fs  => fs                   -- sampling frequency
            )
        port map (
            clk  => clk,
            rst  => rst,
            key  => key,
            gain => gain
            );

    gen_wave : for i in 0 to harmonics-1 generate
        wave_generator_inst : entity work.wave_generator
            generic map (
                G_N  => N-natural(log2(real(harmonics))),
                G_f0 => f0*(natural(i)+1),
                G_fs => fs
                )
            port map (
                I_clk      => clk,
                I_rst      => rst,
                I_wave_sel => wave_sel,
                O_wav      => wave_array(i)
                );
    end generate;

    wave_array_sum(0) <= resize(signed(wave_array(0)), N);
    sum : for i in 1 to harmonics-1 generate
        wave_array_sum(i) <= wave_array_sum(i-1) + resize(shift_right(signed(wave_array(i)), i), N) when config(i) = '1' else wave_array_sum(i-1);  -- divide by 2^i the magnitude of
                                                                                                                                                    -- of the i-th harmonic
                                                                                                                                                    -- of the i-th harmonic
    end generate;

    tone_out_mult <= wave_array_sum(harmonics-1) * signed('0'&gain);

    tone_out <= std_logic_vector(tone_out_mult(N+8 downto 9));
    --tone_out <= std_logic_vector(sine_wave_array_sum(harmonics-1));

end architecture arch;
