library IEEE;
use ieee.numeric_std.all;
use ieee.std_logic_1164.all;
use ieee.math_real.all;

package wave_package is

    type lut_type is array(natural range <>) of signed;

    function sine_table(constant LUT_SIZE, N      : natural; constant f0, fs : real) return lut_type;
    function triangle_table(constant LUT_SIZE, N  : natural; constant f0, fs : real) return lut_type;
    function saw_tooth_table(constant LUT_SIZE, N : natural; constant f0, fs : real) return lut_type;

end wave_package;

package body wave_package is

    function sine_table(constant LUT_SIZE, N : natural; constant f0, fs : real) return lut_type is
        variable sine_lut : lut_type(0 to LUT_SIZE-1)(N-1 downto 0);
        variable sine_int : integer;
    begin
        for i in 0 to LUT_SIZE-1 loop
            sine_int := integer(round(2**(N-1) * sin(MATH_2_PI* f0 * real(i) / fs)));  -- 1 bit for sign, 0 bit for integer part and N-2 bits for fractional part
                                                                                       -- this means that the LUT will contain values in the range -1 to 1-2^(-N+1)
                                                                                       -- and that there will always be an approximation error of 2^(-N+1) in the sine value for +1
            -- saturate the sine value to the range -1 to 1-2^(-N+1)
            if sine_int > 2**(N-1)-1 then
                sine_int := 2**(N-1)-1;
            end if;
            sine_lut(i) := to_signed(sine_int, N);
        end loop;
        return sine_lut;
    end function;


    function triangle_table(constant LUT_SIZE, N : natural; constant f0, fs : real) return lut_type is
        variable triangle_lut : lut_type(0 to LUT_SIZE-1)(N-1 downto 0);
        variable triangle_int : integer;
    begin
        for i in 0 to LUT_SIZE-1 loop
            triangle_int := integer(round(2**(N-1) * real(i) * 4 * f0 / fs));  -- 1 bit for sign, 0 bit for integer part and N-2 bits for fractional part
                                                                               -- this means that the LUT will contain values in the range -1 to 1-2^(-N+1)
                                                                               -- and that there will always be an approximation error of 2^(-N+1) in the value for +1
            -- saturate the sine value to the range -1 to 1-2^(-N+1)
            if triangle_int > 2**(N-1)-1 then
                triangle_int := 2**(N-1)-1;
            end if;
            triangle_lut(i) := to_signed(triangle_int, N);
        end loop;
        return triangle_lut;
    end function;


    function saw_tooth_table(constant LUT_SIZE, N : natural; constant f0, fs : real) return lut_type is
        variable triangle_lut : lut_type(0 to LUT_SIZE-1)(N-1 downto 0);
        variable triangle_int : integer;
    begin
        for i in 0 to LUT_SIZE-1 loop
            triangle_int := integer(round(2**(N-1) * real(i) * 2 * f0 / fs));  -- 1 bit for sign, 0 bit for integer part and N-2 bits for fractional part
                                                                               -- this means that the LUT will contain values in the range -1 to 1-2^(-N+1)
                                                                               -- and that there will always be an approximation error of 2^(-N+1) in the value for +1
            -- saturate the sine value to the range -1 to 1-2^(-N+1)
            if triangle_int > 2**(N-1)-1 then
                triangle_int := 2**(N-1)-1;
            end if;
            triangle_lut(i) := to_signed(triangle_int, N);
        end loop;
        return triangle_lut;
    end function;

end wave_package;
